from __future__ import absolute_import, unicode_literals, print_function
from subprocess import check_output
from prompt_toolkit.completion import Completer, Completion
from fuzzyfinder import fuzzyfinder
import shlex
import json
import os
import os.path
from kubernetes import client, config

class KubectlCompleter(Completer):

    def __init__(self):
        self.all_commands = []
        self.all_args = []
        self.all_opts = []
        self.global_opts = []
        self.inline_help = True
        self.namespace = ""

        try:
            DATA_DIR = os.path.dirname(os.path.realpath(__file__))
            DATA_PATH = os.path.join(DATA_DIR, 'data/cli.json')
            with open(DATA_PATH) as json_file:
                self.kubectl_dict = json.load(json_file)
            self.populate_cmds_args_opts(self.kubectl_dict)
        except Exception as ex:
            print("got an exception" + ex.message)

    def set_inline_help(self, val):
        self.inline_help = val

    def set_namespace(self, namespace):
        self.namespace = namespace

    def populate_cmds_args_opts(self, key_map):
        for key in key_map.keys():
            self.all_commands.append(key)
            self.all_args.extend(key_map[key]['args'])
            if key == "kubectl":
                for opt in key_map[key]['options'].keys():
                    self.global_opts.append(opt)
            else:
                for opt in key_map[key]['options'].keys():
                    self.all_opts.append(opt)

            self.populate_cmds_args_opts(key_map[key]['subcommands'])

    # Below is the grammer of how kubectl expects and parses argument, commands and options
    # kubectl (global option)* command (global option | local option)* (subcommand (global option | local option)*)* (arg) (global option | local option)*
    # This method parse command, args, global flags, local flags from the text before cursor and suggest based on that return state,
    # inner most command seen, arg if any, and pass the dict corresponding to the command
    #
    # states:
    #
    #   INIT: this is starting state. In this state only 'kubectl' is expected as token
    #
    #   KUBECTL: State when first token is 'kubectl' and need to handle second token. In this state only commands of
    #           kubectl or one or more global options can be specified
    #
    #   KUBECTL_CMD: State representing case where we have received 'kubectl' and a 'command'. In this state either args, subcommands
    #           global options, or local options specifc to the commands can be specified
    #
    #   KUBECTL_ARG: In this state only global or local options for the commands can be specified. And also
    #          name os the resource like pods can be specified .We will reach this
    #          state from KUBCTL_CMD when we parse arg for a command
    #
    #   KUBECTL_LEAF: In this state only global or local options for the commands can be specified. We will reach this
    #          state from KUBCTL_CMD when we longer have any sub-commands or args for the previous command
    #
    def parse_tokens(self, cmdline):
        tokens = []
        if cmdline is not None:
            cmdline = cmdline.strip()
            tokens = shlex.split(cmdline)

        state = "INIT"
        index = 0
        key_map = self.kubectl_dict
        last_token = False

        command = ""
        arg = ""
        namespace = self.namespace

        for index, token in enumerate(tokens):

            # if --all-namespaces or --namespace option is passed overide the namespace info from the kubeconfig
            if token == "--all-namespaces":
                namespace = "all"
            if token.startswith("--namespace"):
                if "=" in token:
                    namespace = token.split("=")[1]
            if tokens[index-1] == "--namespace":
                namespace = token

            if state == "INIT" and tokens[0] == "kubectl":
                state = "KUBECTL"
                key_map = key_map['kubectl']
                command = "kubectl"
                continue
            elif state == "KUBECTL":
                if token.startswith("--"):
                    continue
                if token in key_map['subcommands'].keys():
                    state = "KUBECTL_CMD"
                    key_map = key_map['subcommands'][token]
                    command = token
                continue
            elif state == "KUBECTL_CMD":
                if token.startswith("--"):
                    continue
                if token in key_map['subcommands'].keys():
                    state = "KUBECTL_CMD"
                    key_map = key_map['subcommands'][token]
                    command = token
                    continue
                if token in key_map['args']:
                    state = "KUBECTL_ARG"
                    arg = token
                    continue
                continue
            elif state == "KUBECTL_ARG":
                if token.startswith("--"):
                    continue
                resources = self.get_resources(arg)
                if resources:
                    for resource_name, namespace in resources:
                        if token == resource_name:
                            state = "KUBECTL_LEAF"
                            continue
                continue
            elif state == "KUBECTL_LEAF":
                if token.startswith("--"):
                    continue
                continue
        return state, command, arg, key_map, namespace

    def get_completions(self, document, complete_event, smart_completion=None):

        word_before_cursor = document.get_word_before_cursor(WORD=True)

        tokens = []
        cmdline = document.text_before_cursor
        if cmdline is not None:
            cmdline = cmdline.strip()
            try:
                tokens = shlex.split(cmdline)
            except:
                return

        state, command, arg, key_map, namespace = self.parse_tokens(cmdline)

        if state == "INIT":
            self.help_msg = ""
            if self.inline_help:
                self.help_msg = display_meta=key_map['kubectl']['help']
            if len(tokens) == 0:
                yield Completion("kubectl", start_position=0, display="kubectl", display_meta=self.help_msg)
                return
            if len(tokens) == 1 and word_before_cursor == tokens[0]:
                suggestions = fuzzyfinder(tokens[0], ['kubectl'])
                for suggestion in suggestions:
                    yield Completion(suggestion, -len(tokens[0]), display="kubectl", display_meta=self.help_msg)
        elif state == "KUBECTL":
            last_token = tokens[-1]
            self.help_msg = ""
            if word_before_cursor == last_token:
                if last_token.startswith("--"):
                    if last_token in self.global_opts:
                        return
                    global_opt_suggestions = fuzzyfinder(word_before_cursor, self.global_opts)
                    for global_opt in global_opt_suggestions:
                        if self.inline_help:
                            self.help_msg = self.kubectl_dict['kubectl']['options'][global_opt]['help']
                        yield Completion(global_opt, -len(word_before_cursor), display_meta=self.help_msg)
                else:
                    suggestions = fuzzyfinder(last_token, key_map['subcommands'].keys())
                    for suggestion in suggestions:
                        if self.inline_help:
                            self.help_msg = key_map['subcommands'][suggestion]['help']
                        yield Completion(suggestion, -len(last_token), display=suggestion, display_meta=self.help_msg)
            if word_before_cursor == "":
                for cmd in key_map['subcommands'].keys():
                    if self.inline_help:
                        self.help_msg = key_map['subcommands'][cmd]['help']
                    yield Completion(cmd, display=cmd, display_meta=self.help_msg)
        elif state == "KUBECTL_CMD":
            last_token = tokens[-1]
            self.help_msg = ""
            if word_before_cursor == last_token:
                if last_token.startswith("--"):
                    if last_token in self.global_opts or last_token in key_map['options'].keys():
                        return
                    else:
                        command_opts = fuzzyfinder(word_before_cursor, key_map['options'].keys())
                        for command_opt in command_opts:
                            if self.inline_help:
                                self.help_msg = key_map['options'][command_opt]['help']
                            yield Completion(command_opt, -len(word_before_cursor), display=command_opt, display_meta=self.help_msg)
                        global_opt_suggestions = fuzzyfinder(word_before_cursor, self.global_opts)
                        for global_opt in global_opt_suggestions:
                            if self.inline_help:
                                self.help_msg = self.kubectl_dict['kubectl']['options'][global_opt]['help']
                            yield Completion(global_opt, -len(word_before_cursor), display=global_opt, display_meta=self.help_msg)
                elif last_token != command:
                    subcommands = key_map['subcommands'].keys()
                    if len(subcommands) > 0 and last_token not in subcommands:
                        suggestions = fuzzyfinder(last_token, subcommands)
                        for suggestion in suggestions:
                            if self.inline_help:
                                self.help_msg = key_map['subcommands'][suggestion]['help']
                            yield Completion(suggestion, -len(last_token), display=suggestion, display_meta=self.help_msg)
                    args = key_map['args']
                    if len(args) > 0 and last_token not in args:
                        suggestions = fuzzyfinder(last_token, args)
                        for arg in suggestions:
                            yield Completion(arg, -len(last_token))
            elif word_before_cursor == "":
                subcommands = key_map['subcommands'].keys()
                for subcommand in subcommands:
                    if self.inline_help:
                        self.help_msg = key_map['subcommands'][subcommand]['help']
                    yield Completion(subcommand, display=subcommand, display_meta=self.help_msg)
                args = key_map['args']
                for arg in args:
                    yield Completion(arg)
        elif state == "KUBECTL_ARG":
            if word_before_cursor == "":
                resources = self.get_resources(arg, namespace)
                if resources:
                    for resourceName, namespace in resources:
                        yield Completion(resourceName, display=resourceName, display_meta=namespace)
        elif state == "KUBECTL_LEAF":
            last_token = tokens[-1]
            self.help_msg = ""
            if last_token.startswith("--"):
                if last_token in self.global_opts or last_token in key_map['options'].keys():
                    return
                else:
                    command_opts = fuzzyfinder(word_before_cursor, key_map['options'].keys())
                    for command_opt in command_opts:
                        if self.inline_help:
                            self.help_msg = key_map['options'][command_opt]['help']
                        yield Completion(command_opt, -len(word_before_cursor), display=command_opt, display_meta=self.help_msg)
                    global_opt_suggestions = fuzzyfinder(word_before_cursor, self.global_opts)
                    for global_opt in global_opt_suggestions:
                        if self.inline_help:
                            help_msg = self.kubectl_dict['kubectl']['options'][global_opt]['help']
                        yield Completion(global_opt, -len(word_before_cursor), display=global_opt, display_meta=self.help_msg)
        else:
            pass
        return

    def get_resources(self, resource, namespace="all"):
        resources = []
        config.load_kube_config()

        v1 = client.CoreV1Api()
        v1Beta1 = client.AppsV1beta1Api()
        extensionsV1Beta1 = client.ExtensionsV1beta1Api()
        autoscalingV1Api = client.AutoscalingV1Api()
        rbacAPi = client.RbacAuthorizationV1beta1Api()
        batchV1Api = client.BatchV1Api()
        batchV2Api = client.BatchV2alpha1Api()

        ret = None
        namespaced_resource = True

        if resource == "pod":
            ret = v1.list_pod_for_all_namespaces(watch=False)
        elif resource == "service":
            ret = v1.list_service_for_all_namespaces(watch=False)
        elif resource == "deployment":
            ret = v1Beta1.list_deployment_for_all_namespaces(watch=False)
        elif resource == "statefulset":
            ret = v1Beta1.list_stateful_set_for_all_namespaces(watch=False)
        elif resource == "node":
            namespaced_resource = False
            ret = v1.list_node(watch=False)
        elif resource == "namespace":
            namespaced_resource = False
            ret = v1.list_namespace(watch=False)
        elif resource == "daemonset":
            ret = extensionsV1Beta1.list_daemon_set_for_all_namespaces(watch=False)
        elif resource == "networkpolicy":
            ret = extensionsV1Beta1.list_network_policy_for_all_namespaces(watch=False)
        elif resource == "thirdpartyresource":
            namespaced_resource = False
            ret = extensionsV1Beta1.list_third_party_resource(watch=False)
        elif resource == "replicationcontroller":
            ret = v1.list_replication_controller_for_all_namespaces(watch=False)
        elif resource == "replicaset":
            ret = extensionsV1Beta1.list_replica_set_for_all_namespaces(watch=False)
        elif resource == "ingress":
            ret = extensionsV1Beta1.list_ingress_for_all_namespaces(watch=False)
        elif resource == "endpoints":
            ret = v1.list_endpoints_for_all_namespaces(watch=False)
        elif resource == "configmap":
            ret = v1.list_config_map_for_all_namespaces(watch=False)
        elif resource == "event":
            ret = v1.list_event_for_all_namespaces(watch=False)
        elif resource == "limitrange":
            ret = v1.list_limit_range_for_all_namespaces(watch=False)
        elif resource == "configmap":
            ret = v1.list_config_map_for_all_namespaces(watch=False)
        elif resource == "persistentvolume":
            namespaced_resource = False
            ret = v1.list_persistent_volume(watch=False)
        elif resource == "secret":
            ret = v1.list_secret_for_all_namespaces(watch=False)
        elif resource == "resourcequota":
            ret = v1.list_resource_quota_for_all_namespaces(watch=False)
        elif resource == "componentstatus":
            namespaced_resource = False
            ret = v1.list_component_status(watch=False)
        elif resource == "podtemplate":
            ret = v1.list_pod_template_for_all_namespaces(watch=False)
        elif resource == "serviceaccount":
            ret = v1.list_service_account_for_all_namespaces(watch=False)
        elif resource == "horizontalpodautoscaler":
            ret = autoscalingV1Api.list_horizontal_pod_autoscaler_for_all_namespaces(watch=False)
        elif resource == "clusterrole":
            namespaced_resource = False
            ret = rbacAPi.list_cluster_role(watch=False)
        elif resource == "clusterrolebinding":
            namespaced_resource = False
            ret = rbacAPi.list_cluster_role_binding(watch=False)
        elif resource == "job":
            ret = batchV1Api.list_job_for_all_namespaces(watch=False)
        elif resource == "cronjob":
            ret = batchV2Api.list_cron_job_for_all_namespaces(watch=False)
        elif resource == "scheduledjob":
            ret = batchV2Api.list_scheduled_job_for_all_namespaces(watch=False)

        if ret:
            for i in ret.items:
                if namespace == "all" or not namespaced_resource:
                    resources.append((i.metadata.name, i.metadata.namespace))
                elif namespace == i.metadata.namespace:
                    resources.append((i.metadata.name, i.metadata.namespace))
            return resources
        return None




from __future__ import print_function, absolute_import
from pygments.lexer import RegexLexer
from pygments.lexer import words
from pygments.token import Keyword, Name, Operator, Generic, Literal, Text

from kubeshell.completer import KubectlCompleter

class KubectlLexer(RegexLexer):
    """Provides highlighting for commands, subcommands, arguments, and options.

    """
    completer = KubectlCompleter()

    tokens = {
        'root': [
            (words(
                tuple(['kubectl', 'clear', 'exit']),
                prefix=r'\b',
                suffix=r'\b'),
             Literal.String),
            (words(
                tuple(completer.all_args),
                prefix=r'\b',
                suffix=r'\b'),
             Name.Class),
            (words(
                tuple(completer.all_commands),
                prefix=r'\b',
                suffix=r'\b'),
             Name.Class),
            (words(
                tuple(completer.all_opts),
                prefix=r'',
                suffix=r'\b'),
            Keyword),
            (words(
                tuple(completer.global_opts),
                prefix=r'',
                suffix=r'\b'),
            Keyword),
            # Everything else
            (r'.*\n', Text),
        ]
    }



# Copyright 2015 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
from __future__ import print_function, absolute_import, unicode_literals
from pygments.token import Token
from pygments.util import ClassNotFound
from pygments.styles import get_style_by_name
from prompt_toolkit.styles import default_style_extensions, style_from_dict


class StyleFactory(object):
    """Provide styles for the autocomplete menu and the toolbar.

    :type style: :class:`pygments.style.StyleMeta`
    :param style: Contains pygments style info.
    """

    def __init__(self, style_name):
        self.style = self.style_factory(style_name)

    def style_factory(self, style_name):
        """Retrieve the specified pygments style.

        If the specified style is not found, the vim style is returned.

        :type style_name: str
        :param style_name: The pygments style name.

        :rtype: :class:`pygments.style.StyleMeta`
        :return: Pygments style info.
        """
        try:
            style = get_style_by_name(style_name)
        except ClassNotFound:
            style = get_style_by_name('vim')

        # Create a style dictionary.
        styles = {}
        styles.update(style.styles)
        styles.update(default_style_extensions)
        t = Token
        styles.update({
            t.Menu.Completions.Completion.Current: 'bg:#00aaaa #000000',
            t.Menu.Completions.Completion: 'bg:#008888 #ffffff',
            t.Menu.Completions.Meta.Current: 'bg:#00aaaa #000000',
            t.Menu.Completions.Meta: 'bg:#00aaaa #ffffff',
            t.Scrollbar.Button: 'bg:#003333',
            t.Scrollbar: 'bg:#00aaaa',
            t.Toolbar: 'bg:#222222 #cccccc',
            t.Toolbar.Off: 'bg:#222222 #696969',
            t.Toolbar.On: 'bg:#222222 #ffffff',
            t.Toolbar.Search: 'noinherit bold',
            t.Toolbar.Search.Text: 'nobold',
            t.Toolbar.System: 'noinherit bold',
            t.Toolbar.Arg: 'noinherit bold',
            t.Toolbar.Arg.Text: 'nobold'
        })

        return style_from_dict(styles)



from __future__ import print_function, absolute_import, unicode_literals

from prompt_toolkit import prompt
from prompt_toolkit.history import FileHistory
from prompt_toolkit.auto_suggest import AutoSuggestFromHistory
from pygments.style import Style
from pygments.token import Token
from pygments.token import Keyword, Name, Operator, Generic, Literal, Text
from pygments.styles.default import DefaultStyle
from prompt_toolkit.key_binding.defaults import load_key_bindings_for_prompt
from prompt_toolkit.keys import Keys

from kubeshell.style import StyleFactory
from kubeshell.completer import KubectlCompleter
from kubeshell.lexer import KubectlLexer
from kubeshell.toolbar import Toolbar

import os
import click
import sys
import subprocess
import yaml


inline_help=True


registry = load_key_bindings_for_prompt()
completer = KubectlCompleter()

class Kubeshell(object):

    def __init__(self, refresh_resources=True):
        self.history = FileHistory(os.path.expanduser("~/.kube/shell/history"))
        if not os.path.exists(os.path.expanduser("~/.kube/shell/")):
            os.makedirs(os.path.expanduser("~/.kube/shell/"))
        self.toolbar = Toolbar(self.get_cluster_name, self.get_namespace, self.get_user, self.get_inline_help)

        self.clustername = self.user = ""
        self.namespace = "default"

    @registry.add_binding(Keys.F4)
    def _(event):
        global inline_help
        if inline_help:
            inline_help = False
        else:
            inline_help = True
        completer.set_inline_help(inline_help)

    @registry.add_binding(Keys.F10)
    def _(event):
        sys.exit()

    def get_cluster_name(self):
        return self.clustername

    def get_namespace(self):
        return self.namespace

    def get_user(self):
        return self.user

    def get_inline_help(self):
        return inline_help

    def parse_kubeconfig(self):
        self.clustername = self.user = ""
        self.namespace = "default"

        if not os.path.exists(os.path.expanduser("~/.kube/config")):
            return

        with open(os.path.expanduser("~/.kube/config"), "r") as fd:
            docs = yaml.load_all(fd)
            for doc in docs:
                current_context = ""
                for k,v in doc.items():
                    if k == "current-context":
                        current_context = v
                for k,v in doc.items():
                    if k == "contexts":
                        for context in v:
                            if context['name'] == current_context:
                                if 'cluster' in context['context']:
                                    self.clustername = context['context']['cluster']
                                if 'namespace' in context['context']:
                                    self.namespace = context['context']['namespace']
                                if 'user' in context['context']:
                                    self.user = context['context']['user']

    def run_cli(self):

        def get_title():
            return "kube-shell"

        if not os.path.exists(os.path.expanduser("~/.kube/config")):
            click.secho('Kube-shell uses ~/.kube/config for server side completion. Could not find ~/.kube/config. \
                    Server side completion functionality may not work.', fg='red', blink=True, bold=True)
        while True:
            global inline_help
            try:
                self.parse_kubeconfig()
            except:
                # TODO: log errors to log file
                pass
            completer.set_namespace(self.namespace)
            user_input = prompt('kube-shell> ',
                        history=self.history,
                        auto_suggest=AutoSuggestFromHistory(),
                        style=StyleFactory("vim").style,
                        lexer=KubectlLexer,
                        get_title=get_title,
                        enable_history_search=False,
                        get_bottom_toolbar_tokens=self.toolbar.handler,
                        vi_mode=True,
                        key_bindings_registry=registry,
                        completer=completer)
            if user_input == "clear":
                click.clear()
            elif user_input == "exit":
                sys.exit()
            if user_input:
                if '-o' in user_input and 'json' in user_input:
                    user_input = user_input + ' | pygmentize -l json'
                p = subprocess.Popen(user_input, shell=True)
                p.communicate()



from kubeshell import __version__
from setuptools import setup, find_packages

requires = [
    'prompt-toolkit>=1.0.0,<1.1.0',
    'Pygments>=2.1.3,<3.0.0',
    'fuzzyfinder>=1.0.0',
    'click>=4.0,<7.0',
    'kubernetes>=0.10.0',
]

setup(
    name='kube-shell',
    version=__version__,
    description='Kubernetes shell: An integrated shell for working with the Kubernetes CLI',
    author='Cloudnative Labs',
    url='https://github.com/cloudnativelabs/kube-shell',
    packages=find_packages(),
    package_data={'kubeshell': ['data/cli.json']},
    include_package_data=True,
    zip_safe=False,
    install_requires=requires,
    entry_points={
        'console_scripts': [
            'kube-shell = kubeshell.main:cli',
        ]
    },
    license="Apache License 2.0",
    keywords=('kubernetes', 'autocomplete', 'shell',),
    classifiers=(
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'Natural Language :: English',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ),
)



from __future__ import print_function, absolute_import, unicode_literals
from pygments.token import Token
from pygments.token import Keyword, Name, Operator, Generic, Literal, Text

class Toolbar(object):
    """Show information about the aws-shell in a tool bar.

    :type handler: callable
    :param handler: Wraps the callable `get_toolbar_items`.

    """

    def __init__(self, get_cluster_name, get_namespace, get_user, get_inline_help):
        self.handler = self._create_toolbar_handler(get_cluster_name, get_namespace, get_user, get_inline_help)

    def _create_toolbar_handler(self, get_cluster_name, get_namespace, get_user, get_inline_help):
        def get_toolbar_items(_):
            if get_inline_help():
                help_token = Token.Toolbar.On
                help = "ON"
            else:
                help_token = Token.Toolbar.Off
                help = "OFF"

            return [
                (Keyword, ' Cluster: '),
                (Token.Toolbar, get_cluster_name()),
                (Keyword, ' Namespace: '),
                (Token.Toolbar, get_namespace()),
                (Keyword, ' User: '),
                (Token.Toolbar, get_user()),
                (help_token,  ' [F4] In-line help: {0} '.format(help)),
                (Name.Label, ' [F10] Exit ')
            ]

        return get_toolbar_items
